from typing import Dict

from flask import Flask, request, Response
from flask_restful import Resource, Api

from clams.app import ClamsApp

from mmif import Mmif


class Restifier(object):
    """
    Resitifier is a simple wrapper that takes a :class:`.ClamsApp` object and
    turn it into a flaks-based HTTP server. For mapping between Python API and
    HTTP API, it relies on :class:`.ClamsHTTPApi` class.

    Constructor takes a :class:`.ClamsApp` instance and some options for flask configuration.

    :param app_instance: A :class:`.ClamsApp` to wrap.
    :param loopback: when True, the flask wrapper only listens to requests from localhost (used for debugging).
    :param port: Port number for the flask app to listen (used for debugging).
    :param debug: When True, the flask wrapper will run in `debug mode <https://flask.palletsprojects.com/en/1.1.x/quickstart/#debug-mode>`_.
    """
    def __init__(self, app_instance: ClamsApp, loopback: bool = False, port: int = 5000, debug: bool = True) -> None:
        super().__init__()
        self.cla = app_instance
        self.import_name = app_instance.__class__.__name__
        self.flask_app = Flask(self.import_name)
        self.host = 'localhost' if loopback else '0.0.0.0'
        self.port = port
        self.debug = debug
        api = Api(self.flask_app)
        api.add_resource(ClamsHTTPApi, '/',
                         resource_class_args=[self.cla])
    
    def run(self, **options):
        """
        Starts a development server. See :meth:`serve_development`.
        
        :param options: any additional options to pass to the web server.
        """
        self.serve_development(**options)
        
    def serve_production(self, **options):
        """
        Runs the CLAMS app as a flask webapp, using a production-ready web server (gunicorn, https://docs.gunicorn.org/en/stable/#).
        
        :param options: any additional options to pass to the web server.
        """
        import gunicorn.app.base
        import multiprocessing

        def number_of_workers():
            return (multiprocessing.cpu_count() * 2) + 1  # +1 to make sure at least two workers are running
        
        class ProductionApplication(gunicorn.app.base.BaseApplication):

            def __init__(self, app, host, port, **options):
                self.options = {
                    'bind': f'{host}:{port}',
                    'workers': number_of_workers(),
                    'threads': 2,
                }
                self.options.update(options)
                self.application = app
                super().__init__()

            def load_config(self):
                config = {key: value for key, value in self.options.items()
                          if key in self.cfg.settings and value is not None}
                for key, value in config.items():
                    self.cfg.set(key.lower(), value)

            def load(self):
                return self.application

        ProductionApplication(self.flask_app, self.host, self.port, **options).run()

    def serve_development(self, **options):
        """
        Runs the CLAMS app as a flask webapp, using flask built-in development server (https://werkzeug.palletsprojects.com/en/2.0.x/).
        
        :param options: any additional options to pass to the web server.
        """
        self.flask_app.run(host=self.host,
                           port=self.port,
                           debug=self.debug, 
                           **options)

    def test_client(self):
        """
        Returns `flask test client <https://flask.palletsprojects.com/en/1.1.x/testing/>`_.
        """
        return self.flask_app.test_client()


class ClamsHTTPApi(Resource):
    """
    ClamsHTTPApi provides mapping from HTTP verbs to Python API defined in :class:`.ClamsApp`.

    Constructor takes an instance of :class:`.ClamsApp`.
    """
    def __init__(self, cla_instance: ClamsApp):
        super().__init__()
        self.cla = cla_instance
        self.metadata_param_caster = ParameterCaster(self.cla.metadata_param_spec)
        self.annotate_param_caster = ParameterCaster(self.cla.annotate_param_spec)

    @staticmethod
    def json_to_response(json_str: str, status=200) -> Response:
        """
        Helper method to convert JSON output from a ClamsApp to a HTTP response.

        :param json_str: a serialized JSON .
        :param status: a numerical HTTP code to respond.
        :return: A HTTP response ready to send.
        """
        if not isinstance(json_str, str):
            json_str = str(json_str)
        return Response(response=json_str, status=status, mimetype='application/json')

    def get(self) -> Response:
        """
        Maps HTTP GET verb to :func:`~clams.app.ClamsApp.appmetadata`.

        :return: Returns app metadata in a HTTP response.
        """
        return self.json_to_response(self.cla.appmetadata(**self.metadata_param_caster.cast(request.args)))

    def post(self) -> Response:
        """
        Maps HTTP POST verb to :func:`~clams.app.ClamsApp.annotate`.
        Note that for now HTTP PUT verbs is also mapped to :func:`~clams.app.ClamsApp.annotate`.

        :return: Returns MMIF output from a ClamsApp in a HTTP response.
        """
        in_mmif = Mmif(request.get_data())
        params = self.annotate_param_caster.cast(request.args)
        try:
            return self.json_to_response(self.cla.annotate(in_mmif, **params))
        except Exception as e:
            return self.json_to_response(self.cla.record_error(in_mmif, params).serialize(pretty=True), status=500)

    put = post


class ParameterCaster(object):
    """
    A helper class to convert parameters passed by HTTP query strings to
    proper python data types.

    :param param_spec: A specification of a data types of parameters
    """
    def __init__(self, param_spec: Dict):
        self.param_spec = param_spec

    def cast(self, args):
        """
        Given parameter specification, tries to cast values of args to specified Python data types.
        If the parameter key is not specified in the app metadata, a ``TypeError` is raised. 

        :param args: k-v pairs
        :return: A new dictionary of type-casted args
        :raises: KeyError: when parameter is unknown
        """
        casted = {}
        for k, v in args.items():
            if k in self.param_spec:
                if self.param_spec[k] == bool:
                    casted[k] = self.bool_param(v)
                elif self.param_spec[k] == float:
                    casted[k] = self.float_param(v)
                elif self.param_spec[k] == int:
                    casted[k] = self.int_param(v)
                elif self.param_spec[k] == str:
                    casted[k] = self.str_param(v)
                else:
                    casted[k] = v
            else:
                raise KeyError(f"An unknown configuration is passed as a parameter: {k} = {v}")
        return casted

    @staticmethod
    def bool_param(value):
        """
        Helper function to convert string values to bool type.
        """
        return False if value in (False, 0, 'False', 'false', '0') else True

    @staticmethod
    def float_param(value):
        """
        Helper function to convert string values to float type.
        """
        return float(value)

    @staticmethod
    def int_param(value):
        """
        Helper function to convert string values to int type.
        """
        return int(value)

    @staticmethod
    def str_param(value):
        """
        Helper function to convert string values to string type.
        """
        return value



import sys

from mmif import __specver__

from clams import source
from clams.app import *
from clams.app import __all__ as app_all
from clams.appmetadata import AppMetadata
from clams.restify import Restifier
from clams.source import PipelineSource
from clams.ver import __version__

__all__ = [AppMetadata, 'Restifier', 'PipelineSource'] + app_all
version_template = "{} (based on MMIF spec: {})"


def prep_argparser():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-v', '--version',
        action='version',
        version=version_template.format(__version__, __specver__)
    )
    subparsers = parser.add_subparsers(dest='subcmd')
    subparsers.add_parser('source', parents=[source.prep_argparser()], add_help=False)
    return parser


def cli():
    parser = prep_argparser()
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()
    if args.subcmd == 'source':
        source.main(args)



import argparse
import itertools
import json
import sys
from os import path
from typing import Union, Generator, List, Optional, Iterable

from mmif import Mmif, Document, DocumentTypes, __specver__
from mmif.serialize.mmif import MmifMetadata

__all__ = ['PipelineSource']

DOC_JSON = Union[str, dict]
DOC = Union[DOC_JSON, Document]
METADATA_JSON = Union[str, dict]
METADATA = Union[METADATA_JSON, MmifMetadata]


class PipelineSource:
    """
    The PipelineSource class.

    A PipelineSource object is used at the beginning of a
    CLAMS pipeline to populate a new MMIF file with media.

    The same PipelineSource object can be used repeatedly
    to generate multiple MMIF objects.

    :param common_documents_json:
        JSON doc_lists for any documents that should be common
        to all MMIF objects produced by this pipeline.

    :param common_metadata_json:
        JSON doc_lists for metadata that should be common to
        all MMIF objects produced by this pipeline.
    """
    mmif: Mmif

    def __init__(
            self,
            common_documents_json: Optional[List[DOC_JSON]] = None,
            common_metadata_json: Optional[METADATA_JSON] = None
    ) -> None:
        if common_documents_json is None:
            common_documents_json = []
        if common_metadata_json is None:
            common_metadata_json = dict()
        self.mmif_start: dict = {"documents": [json.loads(document)
                                               if isinstance(document, str)
                                               else document
                                               for document in common_documents_json],
                                 "views": [],
                                 "metadata": {
                                     "mmif": f"http://mmif.clams.ai/{__specver__}",
                                     **common_metadata_json
                                 }}
        self.prime()

    def add_document(self, document: Union[str, dict, Document]) -> None:
        """
        Adds a document to the working source MMIF.

        When you're done, fetch the source MMIF with produce().

        :param document: the medium to add, as a JSON dict
                         or string or as a MMIF Medium object
        """
        if isinstance(document, (str, dict)):
            document = Document(document)
        self.mmif.add_document(document)

    def change_metadata(self, key: str, value):
        """
        Adds or changes a metadata entry in the working source MMIF.

        :param key: the desired key of the metadata property
        :param value: the desired value of the metadata property
        """
        self.mmif.metadata[key] = value

    def prime(self) -> None:
        """
        Primes the PipelineSource with a fresh MMIF object.

        Call this method if you want to reset the PipelineSource
        without producing a MMIF object with produce().
        """
        self.mmif = Mmif(self.mmif_start, frozen=False)

    def produce(self) -> Mmif:
        """
        Returns the source MMIF and resets the PipelineSource.

        Call this method once you have added all the documents
        for your pipeline.

        :return: the current MMIF object that has been prepared
        """
        source = self.mmif
        source.freeze_documents()
        self.prime()
        return source

    def __call__(
            self,
            documents: Optional[List[DOC]] = None,
            metadata: Optional[METADATA] = None
    ) -> Mmif:
        """
        Callable API that produces a new MMIF object from
        this pipeline source given a list of documents and
        a metadata object.

        Call with no parameters to produce the default MMIF
        object from ``self.mmif_start``.

        :param documents: a list of additional documents to add
        :param metadata: additional metadata fields to add
        :return: the produced MMIF object
        """
        if documents is None:
            documents = []
        if metadata is None:
            metadata = {}

        if isinstance(documents, str):
            documents = json.loads(documents)
        if isinstance(metadata, MmifMetadata):
            metadata = metadata.serialize() # pytype: disable=attribute-error # bug in pytype? (https://github.com/google/pytype/issues/533)
        if isinstance(metadata, str):
            metadata = json.loads(metadata)

        for document in documents:
            self.add_document(document)
        for key, value in metadata.items():
            self.change_metadata(key, value)
        return self.produce()

    def from_data(
            self,
            doc_lists: Iterable[List[DOC]],
            metadata_objs: Optional[Iterable[Optional[METADATA]]] = None
    ) -> Generator[Mmif, None, None]:
        """
        Provided with an iterable of document lists and an
        optional iterable of metadata objects, generates
        MMIF objects produced from that data.

        ``doc_lists`` and ``metadata_objs`` should be matched pairwise,
        so that if they are zipped together, each pair defines
        a single MMIF object from this pipeline source.

        :param doc_lists: an iterable of document lists to generate MMIF from
        :param metadata_objs: an iterable of metadata objects paired with the document lists
        :return: a generator of produced MMIF files from the data
        """
        if metadata_objs is None:
            metadata_objs = itertools.repeat(None)
        for documents, metadata in zip(doc_lists, metadata_objs):
            yield self(documents, metadata)

    def __iter__(self):
        """
        Endlessly produces MMIF directly from ``self.mmif_start``.

        If called after adding documents or changing metadata,
        these changes are discarded, as the pipeline source
        gets re-primed.
        """
        self.prime()
        while True:
            yield self.produce()


def generate_source_mmif(documents, prefix=None, **ignored):
    from string import Template
    at_types = {
        'video': DocumentTypes.VideoDocument,
        'audio': DocumentTypes.AudioDocument,
        'text': DocumentTypes.TextDocument,
        'image': DocumentTypes.ImageDocument
    }
    template = Template('''{
          "@type": "${at_type}",
          "properties": {
            "id": "${aid}",
            "mime": "${mime}",
            "location": "${location}" }
        }''')
    pl = PipelineSource()
    if prefix and not path.isabs(prefix):
        raise ValueError(f"prefix must be an absolute path; given \"{prefix}\".")

    for doc_id, arg in enumerate(documents, start=1):
        arg = arg.strip()
        if len(arg) < 1:
            continue
        result = arg.split(':', maxsplit=1)
        if len(result) == 2 and result[0].split('/', maxsplit=1)[0] in at_types:
            mime, location = result
        else:
            raise ValueError(
                f'Invalid MIME types, or no MIME type and/or path provided, in argument {doc_id-1} to source'
            )
        if prefix and path.isabs(location):
            raise ValueError(f"when prefix is used, file location must not be an absolute path; given \"{location}\".")
        elif not prefix and not path.isabs(location):
            raise ValueError(f'file location must be an absolute path, or --prefix must be used; given \"{location}\".')
        elif prefix and not path.isabs(location):
            location = path.join(prefix, location)
        doc = template.substitute(
            at_type=str(at_types[mime.split('/', maxsplit=1)[0]]),
            aid=f'd{doc_id}',
            mime=mime,
            location=location
        )
        pl.add_document(doc)
    return pl.produce().serialize(pretty=True)


def prep_argparser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'documents',
        default=None,
        action='store',
        nargs='+',
        help="This list of documents should be colon-joined pairs of document types and file paths. A document type "
             "can be one of ``audio``, ``video``, ``text``, ``image``, or a MIME type string (such as video/mp4). The "
             "output will be a MMIF file containing a document for each of those file paths, with the appropriate "
             "``@type`` and MIME type (if given), printed to the standard output. "
    )
    parser.add_argument(
        '-p', '--prefix',
        default=None,
        metavar='PATH',
        nargs='?',
        help='An absolute path to use as prefix for document file paths. When prefix is set, document file paths MUST '
             'be relative. This can be useful when creating source MMIF files from a system that\'s different from '
             'the system that actually runs the pipeline (e.g. in a docker container).'
    )
    parser.add_argument(
        '-o', '--output',
        default=None,
        action='store',
        nargs='?',
        help='A name of a file to capture a generated MMIF json. When not given, MMIF is printed to stdout.'
    )
    return parser


def main(args):
    if args.output:
        out_f = open(args.output, 'w')
    else:
        out_f = sys.stdout
    out_f.write(generate_source_mmif(**vars(args)))

if __name__ == '__main__':
    parser = prep_argparser()
    args = parser.parse_args()
    main(args)



#! /usr/bin/env python3
import distutils.cmd
import os
from os import path
import shutil

import setuptools

name = "clams-python"
cmdclass = {}

try:
    from sphinx.setup_command import BuildDoc

    cmdclass['build_sphinx'] = BuildDoc
except ImportError:
    print('WARNING: sphinx not available, not building docs')

with open("VERSION", 'r') as version_f:
    version = version_f.read().strip()

with open('README.md') as readme:
    long_desc = readme.read()

with open('requirements.txt') as requirements:
    requires = requirements.readlines()

ver_pack_dir = path.join('clams', 'ver')
shutil.rmtree(ver_pack_dir, ignore_errors=True)
os.makedirs(ver_pack_dir, exist_ok=True)
init_mod = open(path.join(ver_pack_dir, '__init__.py'), 'w')
init_mod.write(f'__version__ = "{version}"')
init_mod.close()


class DoNothing(distutils.cmd.Command):
    description = "run base code until `setuptools.setup()` line and exits 0."
    user_options = []

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self):
        pass


cmdclass['donothing'] = DoNothing

setuptools.setup(
    name="clams-python",
    version=version,
    author="Brandeis Lab for Linguistics and Computation",
    author_email="admin@clams.ai",
    description="A collection of APIs to develop CLAMS app for python",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url="https://clams.ai",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Framework :: Flask',
        'Framework :: Pytest',
        'Intended Audience :: Developers ',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
    ],
    cmdclass=cmdclass,
    command_options={
        'build_sphinx': {
            'source_dir': ('setup.py', 'documentation'),
            'project': ('setup.py', name),
            'version': ('setup.py', version),
            #  'release': ('setup.py', release),
            'build_dir': ('setup.py', 'documentation/_build'),
            'builder': ('setup.py', 'html'),
        }
    },
    install_requires=requires,
    python_requires='>=3.6',
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'clams = clams.__init__:cli',
        ],
    },
)



import os
import sys
from abc import ABC, abstractmethod
from contextlib import contextmanager
from urllib import parse as urlparser

__all__ = ['ClamsApp']

from typing import Union, Any, Optional

from mmif import Mmif, Document, DocumentTypes, View, __specver__
from clams.appmetadata import AppMetadata


class ClamsApp(ABC):
    """
    An abstract class to define API's for ClamsApps. A CLAMS app should inherit
    this class and then can be used with classes in :mod:`.restify` to work as
    web applications.
    """

    def __init__(self):
        self.metadata: AppMetadata = self._appmetadata()
        super().__init__()
        # data type specification for common parameters
        self.metadata_param_spec = {'pretty': bool}
        self.annotate_param_spec = {'pretty': bool}
        python_type = {"boolean": bool, "number": float, "integer": int, "string": str}
        for param_spec in self.metadata.parameters:
            self.annotate_param_spec[param_spec.name] = python_type[param_spec.type]

    def appmetadata(self, **kwargs) -> str:
        """
        A public method to get metadata for this app as a string.

        :return: Serialized JSON string of the metadata
        """
        pretty = kwargs.pop('pretty') if 'pretty' in kwargs else False
        if pretty:
            return self.metadata.json(exclude_defaults=True, by_alias=True, indent=2)
        else:
            return self.metadata.json(exclude_defaults=True, by_alias=True)

    @abstractmethod
    def _appmetadata(self) -> AppMetadata:
        """
        An abstract method to generate (or load if stored elsewhere) the app metadata
        at runtime. All CLAMS app must implement this. For metadata specification, 
        see `https://sdk.clams.ai/appmetadata.jsonschema <../appmetadata.jsonschema>`_. 

        :return: A Python object of the metadata, must be JSON-serializable
        """
        raise NotImplementedError()

    @staticmethod
    def _check_mmif_compatibility(target_specver, input_specver):
        return target_specver.split('.')[:2] == input_specver.split('.')[:2]

    def annotate(self, mmif: Union[str, dict, Mmif], **runtime_params) -> str:
        """
        A public method to invoke the primary app function. It's essentially a
        wrapper around :func:`~clams.app.ClamsApp._annotate` method where some common operations
        (that are invoked by keyword arguments) are implemented.

        :param mmif: An input MMIF object to annotate
        :param runtime_params: An arbitrary set of k-v pairs to configure the app at runtime
        :return: Serialized JSON string of the output of the app
        """
        # TODO (krim @ 12/17/20): add documentation on what are "common" operations
        # should pop all "common" parameters before passing the args to _annotate()
        pretty = runtime_params.pop('pretty') if 'pretty' in runtime_params else False
        if not isinstance(mmif, Mmif):
            mmif = Mmif(mmif)
        input_specver = mmif.metadata.mmif.rsplit('/')[-1]  # pytype: disable=attribute-error
        if 'dev' not in __specver__ :
            if not self._check_mmif_compatibility(__specver__, input_specver):
                raise ValueError(f"Input MMIF file (versioned: {input_specver} is not compatible with the app "
                                 f"targeting at {__specver__}. Make sure apps in the pipeline is all compatible. See "
                                 f"https://mmif.clams.ai/versioning/ for information about MMIF compatibility. ") 
        annotated = self._annotate(mmif, **runtime_params)
        return annotated.serialize(pretty=pretty)

    @abstractmethod
    def _annotate(self, mmif: Union[str, dict, Mmif], **runtime_params) -> Mmif:
        """
        An abstract method to generate (or load if stored elsewhere) the app metadata
        at runtime. All CLAMS app must implement this.

        :param mmif: An input MMIF object to annotate
        :param runtime_params: An arbitrary set of k-v pairs to configure the app at runtime
        :return: A :class:`~mmif.serialize.mmif.Mmif` object of the annotated output, ready for serialization
        """
        raise NotImplementedError()
    
    def get_configuration(self, **runtime_params):
        conf = {}
        for parameter in self.metadata.parameters:
            if parameter.name in runtime_params:
                conf[parameter.name] = str(runtime_params[parameter.name])
            elif parameter.default:
                conf[parameter.name] = parameter.default
            else:
                raise ValueError(f"Cannot find configuration for parameter \"{parameter.name}\".")
        return conf

    def sign_view(self, view: View, runtime_conf: Optional[dict] = None) -> None:
        """
        A method to "sign" a new view that this app creates at the beginning of annotation.
        Signing will populate the view metadata with information and configuration of this app.
        The parameters passed to the :func:`~clams.app.ClamsApp._annotate` must be
        passed to this method. This means all parameters for "common" configuration that
        are consumed in :func:`~clams.app.ClamsApp.annotate` should not be recorded in the
        view metadata.
        :param view: a view to sign
        :param runtime_conf: runtime configuration of the app as k-v pairs
        """
        if view.is_frozen():
            raise ValueError("can't modify an old view")
        view.metadata.app = self.metadata.identifier
        if runtime_conf is not None:
            view.metadata.add_parameters(**{k: str(v) for k, v in runtime_conf.items()})
        
    def set_error_view(self, mmif: Union[str, dict, Mmif], runtime_conf: Optional[dict] = None) -> Mmif:
        """
        A method to record an error instead of annotation results in the view
        this app generated. For logging purpose, the runtime parameters used
        when the error occurred must be passed as well.

        :param mmif: input MMIF object
        :param runtime_conf: parameters passed to annotate when the app encountered the error
        :return: An output MMIF with a new view with the error encoded in the view metadata
        """
        import traceback
        if isinstance(mmif, str) or isinstance(mmif, dict):
            mmif = Mmif(mmif)
        error_view: Optional[View] = None
        for view in reversed(mmif.views):
            if view.metadata.app == self.metadata.identifier:
                error_view = view
                break
        if error_view is None:
            error_view = mmif.new_view()
            self.sign_view(error_view, runtime_conf)
        exc_info = sys.exc_info()
        error_view.set_error(f'{exc_info[0]}: {exc_info[1]}',
                             '\t\n'.join(traceback.format_tb(exc_info[2])))
        return mmif
    
    record_error = set_error_view

    @staticmethod
    def validate_document_locations(mmif: Union[str, Mmif]) -> None:
        """
        Validate files encoded in the input MMIF.

        :param mmif: An input MMIF with zero or more :class:`~mmif.serialize.annotation.Document`
        :raises FileNotFoundError: When any of files is not found at its location
        """
        if isinstance(mmif, str):
            mmif = Mmif(mmif)
        for document in mmif.documents:
            loc = document.location
            if loc is not None and len(loc) > 0:
                p = urlparser.urlparse(loc)
                if p.scheme == 'file':
                    if os.path.exists(p.path):
                        raise FileNotFoundError(f'{document.id}: {loc}')
                # TODO (krim @ 12/15/20): with implementation of file checksum
                #  (https://github.com/clamsproject/mmif/issues/150) , here is a good place for additional check for
                #  file integrity

    @staticmethod
    @contextmanager
    def open_document_location(document: Union[str, Document], opener: Any = open, **openerargs):
        """
        A context-providing file opener. A user can provide their own opener
        class/method and parameters. By default, with will use python built-in
        `open` to open the location of the document.

        :param document: A :class:`~mmif.serialize.annotation.Document` object that has ``location``
        :param opener: A Python class or method that can be used to open a file (e.g. `PIL.Image <https://pillow.readthedocs.io/en/stable/reference/Image.html>`_ for an image file)
        :param openerargs: Parameters that are passed to the ``opener``
        :return:
        """
        if isinstance(document, str):
            document = Document(document)
        if document.location is not None and len(document.location) > 0:
            p = urlparser.urlparse(document.location)
            if p.scheme == 'file':
                if os.path.exists(p.path):
                    if 'mode' not in openerargs and document.at_type == DocumentTypes.TextDocument:
                        openerargs['mode'] = 'r'
                    document_file = opener(p.path, **openerargs)
                    try:
                        yield document_file
                    finally:
                        document_file.close()
                else:
                    raise FileNotFoundError(p.path)



import os
from typing import Union, Dict, List, Optional

import mmif
import pydantic
from mmif import vocabulary
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

primitives = Union[int, float, str, bool]
# these names are taken from the JSON schema data types
param_value_types = Literal['integer', 'number', 'string', 'boolean']

if sys.version_info >= (3, 7):
    param_value_types_values = param_value_types.__args__  # pytype: disable=attribute-error
else:
    param_value_types_values = param_value_types.__values__  # pytype: disable=attribute-error


def get_clams_pyver():
    # real hack to avoid import clams as a package in gh-action workflow
    try:
        import clams
        return clams.__version__
    except ImportError:
        version_fname = os.path.join(os.path.dirname(__file__), '..', '..', 'VERSION')
        if os.path.exists(version_fname):
            with open(version_fname) as version_f:
                return version_f.read().strip()
        else:
            raise Exception('cannot find clams-python version')


def get_mmif_specver():
    return mmif.__specver__


class _BaseModel(pydantic.BaseModel):
    
    class Config:
        def schema_extra(schema, model) -> None:
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)


class Output(_BaseModel):
    """
    Defines a data model that describes output specification of a CLAMS app
    """
    at_type: pydantic.AnyHttpUrl = pydantic.Field(..., alias="@type", description="The type of the object. Must be a IRI string.")
    properties: Dict[str, str] = pydantic.Field({}, description="(optional) Specification for type properties, if any.")
    
    @pydantic.validator('at_type', pre=True)
    def at_type_must_be_str(cls, v):
        if not isinstance(v, str):
            return str(v)
        return v
    
    class Config:
        title = 'CLAMS Output Specification'
        extra = 'forbid'
        allow_population_by_field_name = True

                
class Input(Output):
    """
    Defines a data model that describes input specification of a CLAMS app
    """
    required: bool = pydantic.Field(None, description="(optional) Indicating whether this input type is mandatory or optional.")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.required is None:
            self.required = True

    class Config:
        title = 'CLAMS Input Specification'
        extra = 'forbid'
        allow_population_by_field_name = True


class RuntimeParameter(_BaseModel):
    """
    Defines a data model that describes a single runtime configuration of a CLAMS app. 
    Usually, an app keeps a list of these configuration specifications in the 
    ``parameters`` field. 
    """
    name: str = pydantic.Field(..., description="A short name of the parameter (works as a key).")
    description: str = pydantic.Field(..., description="A longer description of the parameter (what it does, how to use, etc.).")
    type: param_value_types = pydantic.Field(..., description=f"Type of the parameter value the app expects. Must be one of {param_value_types_values}.") 
    choices: List[primitives] = pydantic.Field(None, description="(optional) List of string values that can be accepted.")
    default: primitives = pydantic.Field(None, description="(optional) Default value for the parameter. Only valid for optional parameters. Namely, setting a default value makes a parameter 'optional'. The value must be a string, that is if you want to set a default value for a boolean parameter, be careful to use \'true\' or \'false\' as the value. ")
    
    class Config:
        title = 'CLAMS App Runtime Parameter'
        extra = 'forbid'
        
        
class AppMetadata(pydantic.BaseModel):
    """
    Defines a data model that describes a CLAMS app. 
    Can be initialized by simply passing all required key-value pairs. If you 
    have a pre-generated metadata as an external file, you can read in the file 
    as a ``dict`` and use it as keyword arguments for initialization. 
    
    Please refer to <:ref:`appmetadata`> for the metadata specification. 
    """
    name: str = pydantic.Field(..., description="A short name of the app.")
    description: str = pydantic.Field(..., description="A longer description of the app (what it does, how to use, etc.).")
    app_version: str = pydantic.Field(..., description="Version of the app.")
    mmif_version: str = pydantic.Field(default_factory=get_mmif_specver, description="Version of MMIF specification the app. When the metadata is generated using clams-python SDK, this field is automatically filled in.")
    analyzer_version: str = pydantic.Field(None, description="(optional) Version of an analyzer software, if the app is working as a wrapper for one. ")
    app_license: str = pydantic.Field(..., description="License information of the app.")
    analyzer_license: str = pydantic.Field(None, description="(optional) License information of an analyzer software, if the app is working as a wrapper for one. ")
    identifier: pydantic.AnyHttpUrl = pydantic.Field(..., description="IRI-formatted unique identifier for the app.")
    url: pydantic.AnyHttpUrl = pydantic.Field(..., description="A public repository where the app's source code (git-based) and/or installation specification is available. ")
    input: List[Input] = pydantic.Field([], description="List of input types. Must have at least one.")
    output: List[Output] = pydantic.Field([], description="List of output types. Must have at least one.")
    parameters: List[RuntimeParameter] = pydantic.Field([], description="List of runtime parameters. Can be empty.")
    dependencies: List[str] = pydantic.Field(None, description="(optional) List of software dependencies of the app. This list is completely optional, as in most cases such dependencies are specified in a separate file in the codebase of the app (for example, ``requirements.txt`` file for a Python app, or ``pom.xml`` file for a maven-based Java app). List elements must be strings, not any kind of structured data. Thus, it is recommended to include a package name and its version in the string value at the minimum (e.g., ``clams-python==1.2.3``).")
    more: Dict[str, str] = pydantic.Field(None, description="(optional) A string-to-string map that can be used to store any additional metadata of the app.")

    class Config:
        title = "CLAMS AppMetadata"
        extra = 'forbid'
        allow_population_by_field_name = True

        @staticmethod
        def schema_extra(schema, model) -> None:
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['$schema'] = "http://json-schema.org/draft-07/schema#"  # currently pydantic doesn't natively support the $schema field. See https://github.com/samuelcolvin/pydantic/issues/1478
            schema['$comment'] = f"clams-python SDK {get_clams_pyver()} was used to generate this schema"  # this is only to hold version information
        
    def add_input(self, at_type: Union[str, vocabulary.ThingTypesBase], required: bool = True, **properties):
        """
        Helper method to add an element to the ``input`` list. 
        
        :param at_type: ``@type`` of the input object
        :param required: whether this type is mandatory or optional 
        :param properties: additional property specifications
        """
        new = Input(at_type=at_type, required=required)
        if len(properties) > 0:
            new.properties = properties
        if new not in self.input:
            self.input.append(new)
            if not required:
                # TODO (krim @ 5/12/21): automatically add *optional* input types to parameter
                # see https://github.com/clamsproject/clams-python/issues/29 for discussion
                pass
        else:
            raise ValueError(f"Cannot add a duplicate input '{new}'.")
        
    def add_output(self, at_type: Union[str, vocabulary.ThingTypesBase], **properties):
        """
        Helper method to add an element to the ``output`` list. 
        
        :param at_type: ``@type`` of the input object
        :param properties: additional property specifications
        """
        new = Output(at_type=at_type)
        if len(properties) > 0:
            new.properties = properties
        if new not in self.output:
            self.output.append(new)
        else:
            raise ValueError(f"Cannot add a duplicate output '{new}'.")

    def add_parameter(self, name: str, description: str, type: param_value_types,
                      choices: Optional[List[primitives]] = None, default: primitives = None):
        """
        Helper method to add an element to the ``parameters`` list. 
        """
        new_param = RuntimeParameter(name=name, description=description, type=type, choices=choices, default=default)
        if new_param.name not in [param.name for param in self.parameters]:
            self.parameters.append(new_param)
        else:
            raise ValueError(f"parameter '{new_param.name}' already exist.")
        
    def add_more(self, key:str, value:str):
        """
        Helper method to add a k-v pair to the ``more`` map. 
        :param key: key of an additional metadata
        :param value: value of the additional metadata
        """
        if len(key) > 0 and len(value) > 0: 
            if self.more is None:
                self.more = {}
            if key not in self.more:
                self.more[key] = value
            else:
                raise ValueError(f"'{key}' is already being used in the appmetadata!")
        else:
            raise ValueError("Key and value should not be empty!")
        

if __name__ == '__main__':
    print(AppMetadata.schema_json(indent=2))
